# coding=utf-8


class FeatureExtractor(object):
    def extract(self, instance):
        raise NotImplementedError("{} must implement the 'extract' method".format(self))

    def __init__(self, config):
        raise NotImplementedError("{} must implement the '__init__' method".format(self))

    def batch_extract(self, instances):
        raise NotImplementedError("{} must implement the 'extract_batch' method".format(self))

    def feature_names(self):
        raise NotImplementedError("{} must implement the 'feature_names' method".format(self))



from .. import feature_extraction
import collections
import numpy as np
import os.path
import inspect
import ml.utils as utils
import ml.system as system


class VoiceQualityExtractor(feature_extraction.FeatureExtractor):

    def __init__(self, config):
        self.params = utils.read_config(config)
        # self.voice_quality_config = self.params["voice_quality_config"]
        self.temp_folder = config["temp_folder"] if "temp_folder" in config else "/tmp/opensmile_arffs/"
        system.mkdir_p(self.temp_folder)

    def extract(self, instance):
        last_seconds_values = self.params["extract_on_last_seconds"]

        duration = instance.audio.duration_seconds
        all_values = {}

        all_values["mean_shimmer"] = collections.defaultdict(lambda: np.nan)
        all_values["mean_jitter"] = collections.defaultdict(lambda: np.nan)
        all_values["mean_nhr"] = collections.defaultdict(lambda: np.nan)

        for last_secs in last_seconds_values:
            if last_secs == "all":
                last_secs = duration
                label = "all"
            else:
                label = last_secs

            if duration < last_secs:
                continue

            jitter, shimmer, nhr = self.extract_voice_quality(instance, start=duration - last_secs, end=duration)
            all_values["mean_jitter"][label] = jitter
            all_values["mean_shimmer"][label] = shimmer
            all_values["mean_nhr"][label] = nhr

        feat = {}

        for last_secs in last_seconds_values:
            if last_secs == "all":
                in_ms = "all"
            else:
                in_ms = "last_{}ms".format(int(last_secs * 1000))

            for feat_name in ["mean_jitter", "mean_shimmer", "mean_nhr"]:
                feat["{}_{}".format(feat_name, in_ms)] = all_values[feat_name][last_secs]

        return feat

    def extract_voice_quality(self, instance, start, end):
        min_pitch = 50 if instance.speaker_male() else 75
        max_pitch = 300 if instance.speaker_male() else 500

        script_path = os.path.dirname(inspect.getfile(inspect.currentframe())) + "/voice-analysis.praat"
        try:
            output = system.run_external_command("praat {}".format(script_path), non_named_params=[instance.filename, start, end, min_pitch, max_pitch])
        except:
            system.warning("Check this case (Extracting voice quality error)!!!!")  # TODO: fix
            return np.nan, np.nan, np.nan

        values = dict([o.split(":") for o in output.split()])

        jitter = float(values["sound_voiced_local_jitter"]) if "undefined" not in values["sound_voiced_local_jitter"] else np.nan
        shimmer = float(values["sound_voiced_local_shimmer"]) if "undefined" not in values["sound_voiced_local_shimmer"] else np.nan
        nhr = float(values["noise_to_harmonics_ratio"]) if "undefined" not in values["noise_to_harmonics_ratio"] else np.nan

        return jitter, shimmer, nhr

    def batch_extract(self, instances):
        return [self.extract(instance) for instance in instances]



from .eeg_feature_extractor import EEGFeatureExtractor
import numpy as np
import ml.data_import as data_import
import ml.signal_processing as signal_processing
import ml.system


class WindowedExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(WindowedExtractor, self).__init__(config)
        self.window_sizes = self.params["window_sizes"]
        frame_duration = 1.0 / self.sfreq

        if "step_in_frames" in self.params:
            self.step_in_frames = self.params["step_in_frames"]
            self.step_in_secs = self.params["step_in_frames"] * frame_duration

        elif "step_in_secs" in self.params:
            self.step_in_secs = self.params["step_in_secs"]
            self.step_in_frames = int(self.step_in_secs / frame_duration)
        else:
            raise Exception("missing step parameter")
        ml.system.info("windowed extractor: {} every {} frames ({} ms)".format(["{} ms".format(s*1000.0) for s in self.window_sizes], self.step_in_frames, 1000.0 * self.step_in_secs))

    def extract(self, trial):
        ml.system.info("Trial shape: (channels x samples) {}".format(trial.shape))
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        assert self.n_samples == trial.shape[1], "the number of samples in the extractor configuration doen't match {} vs {}".format(self.n_samples, trial.shape[1])

        features = []
        for window_size_in_secs in self.window_sizes:
            ws_frames = int(window_size_in_secs * self.sfreq)
            if self.verbose:
                print(("ws frames: ", ws_frames))
            for starting_sample, chunk in signal_processing.sliding_window(trial,
                                                                           window_size_in_frames=ws_frames,
                                                                           step_in_frames=self.step_in_frames,
                                                                           time_limits=self.extraction_time_limits,
                                                                           t0=self.trial_tmin,
                                                                           freq=self.sfreq,
                                                                           debugging=self.verbose,
                                                                           ):
                window_coefficients = []
                for ch in self.channels_to_extract_from:
                    y = chunk[ch, :]
                    window_coefficients.append(y.mean())
                features.extend(window_coefficients)

        features = np.array(features)
        return features

    def mapping(self):
        i = 0
        dummy_epoch = np.zeros((self.n_channels, self.n_samples))  # (channels x samples)
        montage = data_import.read_montage(self.params["montage"])

        res = {}
        for window_size_in_secs in self.window_sizes:
            ws_frames = int(window_size_in_secs * self.sfreq)
            for starting_sample, chunk in signal_processing.sliding_window(dummy_epoch,
                                                                           window_size_in_frames=ws_frames,
                                                                           step_in_frames=self.step_in_frames,
                                                                           time_limits=self.extraction_time_limits,
                                                                           t0=self.trial_tmin,
                                                                           freq=self.sfreq,
                                                                           ):
                for ch in self.channels_to_extract_from:
                    t0, t1 = signal_processing.frame_time_limits(starting_sample, self.sfreq, self.trial_tmin, ws_frames)
                    res[i] = dict(
                        channel=ch,
                        channel_name=montage.ch_names[ch],
                        starting_sample=starting_sample,
                        window_size=window_size_in_secs,
                        starting_time=round(t0, 3),
                        end_time=round(t1, 3),
                        channel_position=montage.pos[ch],
                        name="(w{}|c{}|{}|{})".format(round(window_size_in_secs, 2), ch, round(t0, 2), round(t1, 2))
                    )
                    i = i + 1
        return res

    def feature_names(self):
        return [c["name"] for c in list(self.mapping().values())]



# coding=utf-8

import ml.system as system
from . import textgrid


class Intervals():
    def __init__(self, intervals):
        if len(intervals) > 0 and (type(intervals[0]) is tuple or type(intervals[0]) is list):
            intervals = [textgrid.interval(t[0], t[1], t[2]) for t in intervals]

        self.intervals = intervals
        self.__reset__()

    def __reset__(self):
        self.counter = 0

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self.intervals == other.intervals)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        for i in self.intervals:
            yield i

    def __getitem__(self, idx):
        return list(self)[idx]

    def __len__(self):
        return len(self.intervals)

    def validate_values(self, values):
        warnings = []
        for idx, i in enumerate(self.intervals):
            if i.mark not in values:
                warnings.append("(id: {}, time: {}) invalid label \"{}\"".format(idx + 1, i.minTime, i.mark))
        return warnings

    def validate_for_diff(self, x):
        warnings = []
        if len(self.intervals) != len(x.intervals):
            warnings.append("number of intervals: {} vs {}".format(len(self.intervals), len(x.intervals)))
        for i in x:
            if i.minTime != self.current().minTime:
                warnings.append("different starts on interval ({}) {} vs {}".format(self.counter + 1, i.minTime, self.current().minTime))
                break
            next(self)

        self.__reset__()

        return warnings

    def diff(self, x):
        differences = []
        if len(self.intervals) != len(x.intervals):
            system.warning("number of intervals: {} vs {}".format(len(self.intervals), len(x.intervals)))

        for i in x:
            if i.minTime != self.current().minTime:
                system.warning("different starts on interval ({}) {} vs {}".format(self.counter + 1, i.minTime, self.current().minTime))
                break
            if i.mark.strip() != self.current().mark.strip():
                differences.append((self.counter + 1, i.minTime, i.maxTime, i.mark, self.current().mark))
            elif i.mark != self.current().mark:
                system.warning("Labels '{}' vs '{}' ignored from diff".format(i.mark, self.current().mark))

            next(self)

        self.__reset__()

        return differences

    def merge(self, x):
        if len(self.intervals) != len(x.intervals):
            system.warning("number of intervals: {} vs {}".format(len(self.intervals), len(x.intervals)))

        res = []

        for i in x:
            if i.minTime != self.current().minTime:
                system.error("different starts on interval ({}) {} vs {}".format(self.counter + 1, i.minTime, self.current().minTime))
                raise Exception("Intervals are not the same")

            if i.mark == self.current().mark:
                res.append(i)
            else:
                res.append(textgrid.Interval(i.minTime, i.maxTime, "A"))

            next(self)

        self.__reset__()
        return Intervals(res)

    def current(self):
        return self.intervals[self.counter]

    def next(self, value=None):
        if self.has_next():
            self.counter += 1
            val = self.intervals[self.counter]
            if value and value != val.mark:
                return self.next(value)
            else:
                return val
        else:
            return None

    def values_in_range(self, start, end):
        values = []
        for i in self.intervals:
            if i.maxTime < start:
                continue
            elif i.minTime > end:
                break
            else:
                values.append(i.mark)
        return values

    def next_until(self, time):
        if self.current().end < time and self.has_next():
            next(self)
            return self.next_until(time)
        else:
            return self.current()

    def has_next(self):
        return self.counter < len(self.intervals) - 1

    def __repr__(self):
        return "Wavesurfer: " + str(self.intervals)

    def as_vad(self):
        vad_intervals = []
        current_ipu = None
        for i in self:
            if i.mark == "#":
                if current_ipu:
                    vad_intervals.append(current_ipu)
                    current_ipu = None
                vad_intervals.append(textgrid.Interval(i.minTime, i.maxTime, "0"))

            else:
                if not current_ipu:
                    current_ipu = textgrid.Interval(i.minTime, i.maxTime, "1")
                else:
                    current_ipu = textgrid.Interval(current_ipu.minTime, i.maxTime, "1")

        return Intervals(vad_intervals)

    def as_list(self):
        return [(i.minTime, i.maxTime, i.mark) for i in self]



from unittest import TestCase

import os
import ml.utils
import ml.feature_extraction.extractors
import ml.audio_instance
import ml.signal_processing
import ml.system

# ml.system.mute()

class TestAcousticExtractor(TestCase):
    def test_can_extract_from_simple_instance(self):
        config = dict(extraction_method = "acustics",
                        smile_extract_path="/home/pbrusco/apps/opensmile-2.3.0/bin/linux_x64_standalone_static/",
                        pitch_config="/home/pbrusco/apps/opensmile-2.3.0/config/smileF0.conf",
                        intensity_config="/home/pbrusco/apps/opensmile-2.3.0/config/intensity.conf",
                        voice_quality_config="/home/pbrusco/apps/opensmile-2.3.0/config/voice_quality.config",
                        extended_features=False,
                        extract_on_last_seconds=[0.2, 0.3, 0.5]
                     )
        extractor = ml.feature_extraction.extractors.create(config)
        here = os.path.dirname(os.path.abspath(__file__))
        filename = "{}/test.wav".format(here)
        audio = ml.signal_processing.load_wav(filename)
        instance = ml.audio_instance.AudioInstance(filename=filename, speaker_gender="M", audio=audio)
        features = extractor.extract(instance)

        self.assertAlmostEqual(features['f0_slope_last_200ms'], -172.45, delta=0.01)
        self.assertAlmostEqual(features['f0_slope_last_300ms'], -238.14, delta=0.01)
        self.assertAlmostEqual(features['f0_slope_last_500ms'], -19.20, delta=0.01)
        self.assertAlmostEqual(features['ipu_duration'], 2.33, delta=0.01)
        self.assertAlmostEqual(features['mean_intensity_last_200ms'], 65.14, delta=0.01)
        self.assertAlmostEqual(features['mean_intensity_last_300ms'], 65.24, delta=0.01)
        self.assertAlmostEqual(features['mean_intensity_last_500ms'], 63.62, delta=0.01)
        self.assertAlmostEqual(features['mean_jitter_last_200ms'], 0.01, delta=0.01)
        self.assertAlmostEqual(features['mean_jitter_last_300ms'], 0.01, delta=0.01)
        self.assertAlmostEqual(features['mean_jitter_last_500ms'], 0.01, delta=0.01)
        self.assertAlmostEqual(features['mean_nhr_last_200ms'], 0.19, delta=0.01)
        self.assertAlmostEqual(features['mean_nhr_last_300ms'], 0.15, delta=0.01)
        self.assertAlmostEqual(features['mean_nhr_last_500ms'], 0.18, delta=0.01)
        self.assertAlmostEqual(features['mean_pitch_last_200ms'], 159.95, delta=0.01)
        self.assertAlmostEqual(features['mean_pitch_last_300ms'], 170.05, delta=0.01)
        self.assertAlmostEqual(features['mean_pitch_last_500ms'], 168.00, delta=0.01)
        self.assertAlmostEqual(features['mean_shimmer_last_200ms'], 0.25, delta=0.01)
        self.assertAlmostEqual(features['mean_shimmer_last_300ms'], 0.08, delta=0.01)
        self.assertAlmostEqual(features['mean_shimmer_last_500ms'], 0.18, delta=0.01)

    def test_can_extract_from_multiple_instances(self):
        config = dict(extraction_method = "acustics",
                        smile_extract_path="/home/pbrusco/apps/opensmile-2.3.0/bin/linux_x64_standalone_static/",
                        pitch_config="/home/pbrusco/apps/opensmile-2.3.0/config/smileF0.conf",
                        intensity_config="/home/pbrusco/apps/opensmile-2.3.0/config/intensity.conf",
                        voice_quality_config="/home/pbrusco/apps/opensmile-2.3.0/config/voice_quality.config",
                        extended_features=False,
                        extract_on_last_seconds=[0.2, "all"]
                     )
        extractor = ml.feature_extraction.extractors.create(config)
        here = os.path.dirname(os.path.abspath(__file__))

        filename = "{}/test.wav".format(here)
        audio = ml.signal_processing.load_wav(filename)
        instance = ml.audio_instance.AudioInstance(filename=filename, speaker_gender="M", audio=audio)

        filename2 = "{}/test.wav".format(here)
        audio2 = ml.signal_processing.load_wav(filename2)
        instance2 = ml.audio_instance.AudioInstance(filename=filename2, speaker_gender="M", audio=audio2)

        all_features = extractor.batch_extract([instance, instance2])

        self.assertAlmostEqual(all_features[0]['f0_slope_last_200ms'], -172.45, delta=0.01)
        self.assertAlmostEqual(all_features[0]['ipu_duration'], 2.33, delta=0.01)
        self.assertAlmostEqual(all_features[0]['mean_intensity_last_200ms'], 65.14, delta=0.01)
        self.assertAlmostEqual(all_features[0]['mean_jitter_last_200ms'], 0.01, delta=0.01)
        self.assertAlmostEqual(all_features[0]['mean_nhr_last_200ms'], 0.19, delta=0.01)
        self.assertAlmostEqual(all_features[0]['mean_pitch_last_200ms'], 159.95, delta=0.01)
        self.assertAlmostEqual(all_features[0]['mean_shimmer_last_200ms'], 0.25, delta=0.01)



import numpy as np


def get_column(data, column_name):
    column_id = [a[0] for a in data["attributes"]].index(column_name)
    values = np.array([float(d[column_id]) for d in data["data"]])
    return values



import arff
import time
import os
from . import system


def call_script(smile_extract_path, temp_folder, config, filename):
    timestamp = time.time()
    temp_output = "{}/{}{}.arff".format(temp_folder, timestamp, os.path.basename(filename))
    data = {}
    if not system.exists(config):
        raise Exception("{} not found".format(config))
    system.run_command("{}/SMILExtract -C {} -I {} -l 0 -arffoutput {} -appendarff 0".format(smile_extract_path, config, filename, temp_output))
    data = arff.load(open(temp_output))
    system.rm(temp_output)
    return data



# coding: utf-8



import numpy as np
import sys
import configparser
import collections
from os.path import expanduser
from . import system
import csv
import random

HOME = expanduser("~")


def read_config(config_input, require=[], optional=[]):
    if isinstance(config_input, str):
        if not system.exists(config_input):
            raise Exception("missing config: {}".format(config_input))
        config = configparser.SafeConfigParser()
        config.read([config_input])
        res = {}
        for k, v in config.items("DEFAULT"):
            res[k] = eval(v)
    elif isinstance(config_input, dict):
        res = config_input
    else:
        raise Exception("config type not allowed (type={})".format(type(config_input)))

    missing_fields = [f for f in require if f not in res]

    if len(missing_fields) > 0:
        system.error("Missing fields in config:\n\t\t" + "\n\t\t".join(missing_fields))
        raise Exception("Missing required fields configuration")

    for opt in optional:
        if opt not in res:
            res[opt] = None

    return res


def read_list(filename):
    if not system.exists(filename):
        raise Exception("missing list: {}".format(filename))

    file = open(filename, "r")

    res = []
    lines = file.readlines()
    for line in lines:
        if line.strip() != "":
            l = line.replace("~/", HOME + "/")
            chunks = l.split()
            if len(chunks) > 1:
                res.append(chunks)
            else:
                res.append(chunks[0])

    return list(res)


def read_wavesurfer(filename):
    return [(float(line[0]), float(line[1]), " ".join(line[2:])) for line in read_list(filename)]


def read_ipus(vad_filename, zero_tag = "0"):
    return [(t0, tf) for (t0, tf, lbl) in read_wavesurfer(vad_filename) if lbl != zero_tag]


def read_turns(turns_filename):
    return [(t0, tf, tt) for (t0, tf, tt) in read_wavesurfer(turns_filename) if tt != "#"]


def read_csv(filename, delimiter=","):
    res = []
    with open(filename) as csvfile:
        spamreader = csv.reader(csvfile, delimiter=delimiter)
        for i, row in enumerate(spamreader):
            if i == 0:
                header = row
            else:
                res.append(dict(list(zip(header, row))))
        return res


def read_dat_table(filename, header_filename):
    with open(header_filename, "r") as header_file:
        content = header_file.read()
        header = [h.split()[0] for h in content.strip().split("\n")]

    data = []
    with open(filename, "r") as data_file:
        for line in data_file:
            row = dict(list(zip(header, line.strip().split("\t"))))
            data.append(row)
    return data


def absolute_to_user_path(filename):
    return filename.replace(HOME + "/", "~/")


def save_list(lines, filename, verbose=True, separator="\t"):
    with open(filename, "w") as fn:
        for line in lines:
            if type(line) is list or type(line) is tuple:
                fn.write(separator.join([str(l) for l in line]) + "\n")
            else:
                fn.write(line + "\n")
    if verbose:
        system.info("output list saved at {}".format(filename))


def print_inline(str):
    delete = "\b" * (len(str) + 2)
    txt = "{0}{1}".format(delete, str)

    print(txt, end=" ")

    sys.stdout.flush()


def call_with_necesary_params_only(fn, params):
    arg_count = fn.__code__.co_argcount
    args = fn.__code__.co_varnames[:arg_count]

    args_dict = {}
    for k, v in list(params.items()):
        if k in args:
            args_dict[k] = v

    return fn(**args_dict)


def extend_dict(dictionary, key, new_data):
    if key not in dictionary:
        dictionary[key] = new_data
    else:
        dictionary[key] = np.concatenate((dictionary[key], new_data), axis=2)
    return dictionary


def get_param_or(args, arg_number, default):
    if len(args) > arg_number:
        return args[arg_number]
    else:
        return default


def count_if(f, values):
    return len([0 for v in values if f(v)])


def apply_mapping_to_data(X, y, ids, categories_mapping):
    mapped_y = np.array([categories_mapping[y_i] for y_i in y])

    X_filtered = X[mapped_y != np.array(None), :]
    ids_filtered = ids[mapped_y != np.array(None)]
    y_filtered = mapped_y[mapped_y != np.array(None)].astype(int)

    return X_filtered, y_filtered, ids_filtered


def compare_intersection(y_1, y_2):
    intersection = sum([1 if y else 0 for y in (y_1 == y_2)])
    return intersection


def subsample_ids(labels, shuffle=True):
    assert(isinstance(labels, np.ndarray))

    counts = collections.Counter(labels)
    min_count = min(counts.values())

    res = []

    for classs in counts:
        indices = np.arange(len(labels))[labels == classs]
        if shuffle:
            random.shuffle(indices)
        res.extend(indices[:min_count])

    res = np.array(res)
    return res


def unzip(arr):
    lst1, lst2 = list(zip(*arr))
    return list(lst1), list(lst2)


def flatten(list_of_lists):
    return [e for l in list_of_lists for e in l]


def all_equal(l):
    l = np.array(l)
    return l.size == 0 or np.all(l == l[0])



from .eeg_feature_extractor import EEGFeatureExtractor


class WaveletsExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(WaveletsExtractor, self).__init__(config)

    def extract(self, trial):
        pass



# coding=utf-8

import ml.system as system
from . import textgrid # pip install git+http://github.com/kylebgorman/textgrid.git


def read(textgrid_file):
    if not system.exists(textgrid_file):
        raise Exception("Missing file: " + textgrid_file)

    return textgrid.TextGrid().read(textgrid_file)



# coding=utf-8


class EqualityMixin(object):

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)



# coding: utf-8
from .. import feature_extraction
import ml.feature_extraction.audio.pitch_extractor
import ml.feature_extraction.audio.intensity_extractor
import ml.feature_extraction.audio.voice_quality_extractor

import ml.parsing.arff
import ml.opensmile


class AcousticsExtractor(feature_extraction.FeatureExtractor):
    def __init__(self, config):
        self.pitch_extractor = ml.feature_extraction.audio.pitch_extractor.PitchExtractor(config)
        self.intensity_extractor = ml.feature_extraction.audio.intensity_extractor.IntensityExtractor(config)
        self.voice_quality_extractor = ml.feature_extraction.audio.voice_quality_extractor.VoiceQualityExtractor(config)

    def extract(self, instance):
        duration = instance.audio.duration_seconds

        feat = dict(ipu_duration=duration)
        feat.update(self.pitch_extractor.extract(instance))
        feat.update(self.intensity_extractor.extract(instance))
        feat.update(self.voice_quality_extractor.extract(instance))
        return feat

    def _dict_union_(self, *dicts):
        return dict(pair for d in dicts for pair in d.items())

    def batch_extract(self, instances):
        durations = [dict(ipu_duration=i.audio.duration_seconds) for i in instances]
        features1 = self.pitch_extractor.batch_extract(instances)
        features2 = self.intensity_extractor.batch_extract(instances)
        features3 = self.voice_quality_extractor.batch_extract(instances)
        return [self._dict_union_(durations[i], features1[i], features2[i], features3[i]) for i in range(0, len(instances))]



# coding=utf-8



from .. import utils


def create(config):
    params = utils.read_config(config)
    if params["extraction_method"] == "raw":
        import ml.feature_extraction.eeg.raw_extractor
        return ml.feature_extraction.eeg.raw_extractor.RawExtractor(params)
    elif params["extraction_method"] == "freqs":
        import ml.feature_extraction.eeg.freq_extractor
        return ml.feature_extraction.eeg.freq_extractor.FreqExtractor(params)
    elif params["extraction_method"] == "trash":
        import ml.feature_extraction.eeg.trash_extractor
        return ml.feature_extraction.eeg.trash_extractor.TrashExtractor(params)
    elif params["extraction_method"] == "window_mean":
        import ml.feature_extraction.eeg.windowed_extractor
        return ml.feature_extraction.eeg.windowed_extractor.WindowedExtractor(params)
    elif params["extraction_method"] == "wavelets":
        return ml.feature_extraction.eeg.wavelets_extractor.WaveletsExtractor(params)
    elif params["extraction_method"] == "acustics":
        import ml.feature_extraction.audio.acoustics_extractor
        return ml.feature_extraction.audio.acoustics_extractor.AcousticsExtractor(params)



from .. import feature_extraction
import collections
import numpy as np
import ml.utils as utils
import ml.system as system
import ml.opensmile
import ml.parsing.arff


class IntensityExtractor(feature_extraction.FeatureExtractor):

    def __init__(self, config):
        self.params = utils.read_config(config, require=["intensity_config"])
        # self.voice_quality_config = self.params["voice_quality_config"]
        self.temp_folder = config["temp_folder"] if "temp_folder" in config else "/tmp/opensmile_arffs/"
        system.mkdir_p(self.temp_folder)

    def extract(self, instance):
        last_seconds_values = self.params["extract_on_last_seconds"]

        duration = instance.audio.duration_seconds
        times_intensities, intensities = self.intensity_track(instance)

        all_values = {}

        all_values["mean_intensity"] = collections.defaultdict(lambda: np.nan)

        for last_secs in last_seconds_values:
            if last_secs == "all":
                last_secs = duration
                label = "all"
            else:
                label = last_secs

            if duration < last_secs:
                continue

            indices = (times_intensities >= duration - last_secs)
            all_values["mean_intensity"][label] = np.mean(intensities[indices])

        feat = {}

        for last_secs in last_seconds_values:
            if last_secs == "all":
                in_ms = "all"
            else:
                in_ms = "last_{}ms".format(int(last_secs * 1000))

            feat["mean_intensity_{}".format(in_ms)] = all_values["mean_intensity"][last_secs]

        if self.params["extended_features"]:
            feat["intensities"] = intensities
            feat["times_intensities"] = times_intensities

        return feat

    def intensity_track(self, instance):
        data = ml.opensmile.call_script(self.params["smile_extract_path"], self.temp_folder, self.params["intensity_config"], instance.filename)

        times = ml.parsing.arff.get_column(data, "frameTime")
        intensities = ml.parsing.arff.get_column(data, "pcm_intensity_sma")

        times = times[intensities != 0]
        intensities = intensities[intensities != 0]
        intensities_in_dbs = 10 * np.log10(intensities / 10e-14)  # TODO: revisar using I0 = 10 −6 as defined in the opensmile book
        return times, intensities_in_dbs

    def batch_extract(self, instances):
        return [self.extract(instance) for instance in instances]



# coding=utf-8


from .. import feature_extraction

import ml.utils as utils
import numpy as np


class EEGFeatureExtractor(feature_extraction.FeatureExtractor):
    def __init__(self, config):
        self.params = utils.read_config(config, require=["trial_tmin", "trial_tmax", "sfreq", "channels_to_extract_from", "extraction_tmin", "extraction_tmax"])

        self.trial_tmin = float(self.params["trial_tmin"])
        self.trial_tmax = float(self.params["trial_tmax"])
        self.sfreq = float(self.params["sfreq"])
        self.n_samples = int((self.trial_tmax - self.trial_tmin) * self.sfreq)
        self.n_channels = self.params["n_channels"]

        self.extraction_tmin = self.params["extraction_tmin"]
        self.extraction_tmax = self.params["extraction_tmax"]
        self.extraction_time_limits = [self.extraction_tmin, self.extraction_tmax]

        self.channels_to_extract_from = np.array(self.params["channels_to_extract_from"])
        self.verbose = self.params["verbose"] if ("verbose" in self.params) else None

        self.validate_extractor()

        if self.verbose:
            print("extractor params")
            print((self.__dict__))

    def validate_extractor(self):
        msg = "The numbers of channel to extract from are not in range"
        assert np.all((self.channels_to_extract_from >= 0) & (self.channels_to_extract_from < self.n_channels)), msg

        msg = "times are not in order"
        assert self.trial_tmin < self.trial_tmax, msg

        msg = "extraction time is not valid (should be contained in [tmin, tmax] range)"
        in_range = self.extraction_tmin >= self.trial_tmin and \
                   self.extraction_tmin < self.trial_tmax and \
                   self.extraction_tmax > self.trial_tmin and \
                   self.extraction_tmax <= self.trial_tmax and \
                   self.extraction_tmin < self.extraction_tmax

        assert in_range, msg



from .. import feature_extraction
import collections
import pysptk
import numpy as np
import scipy.stats
import ml.utils as utils
import ml.system as system
import ml.opensmile
import ml.parsing.arff
import math


class PitchExtractor(feature_extraction.FeatureExtractor):
    def __init__(self, config):
        self.params = utils.read_config(config, require=["pitch_config"])
        self.temp_folder = config["temp_folder"] if "temp_folder" in config else "/tmp/opensmile_arffs/"
        system.mkdir_p(self.temp_folder)

    def extract(self, instance):
        last_seconds_values = self.params["extract_on_last_seconds"]

        duration = instance.audio.duration_seconds
        times_pitch, pitch = self.pitch_track(instance)

        all_values = {}

        all_values["f0_slope"] = collections.defaultdict(lambda: np.nan)
        all_values["mean_pitch"] = collections.defaultdict(lambda: np.nan)

        for last_secs in last_seconds_values:
            if last_secs == "all":
                last_secs = duration
                label = "all"
            else:
                label = last_secs

            if duration < last_secs:
                continue

            indices = (times_pitch >= duration - last_secs)
            all_values["mean_pitch"][label] = np.mean(pitch[indices])

            indices = (times_pitch >= duration - last_secs)
            if len(times_pitch[indices]) > 3:
                all_values["f0_slope"][label] = scipy.stats.linregress(times_pitch[indices], pitch[indices])[0]  # [0] => slope (m)

        feat = {}
        for last_secs in last_seconds_values:
            if last_secs == "all":
                in_ms = "all"
            else:
                in_ms = "last_{}ms".format(int(last_secs * 1000))

            for feat_name in ["f0_slope", "mean_pitch"]:
                feat["{}_{}".format(feat_name, in_ms)] = all_values[feat_name][last_secs]

        if self.params["extended_features"]:
            feat["pitch"] = pitch
            feat["times_pitch"] = times_pitch

        return feat

    def pitch_track(self, instance):
        data = ml.opensmile.call_script(self.params["smile_extract_path"], self.temp_folder, self.params["pitch_config"], instance.filename)

        times = ml.parsing.arff.get_column(data, "frameTime")
        pitch = ml.parsing.arff.get_column(data, "F0final_sma")

        times = times[pitch > 0]
        pitch = pitch[pitch > 0]

        return times, pitch
        #
        # fs = instance.audio.frame_rate
        # ms_10_in_frames = math.ceil(fs / 100) # Extract every 10 ms.
        # min_pitch = 50 if instance.speaker_male() else 75
        # max_pitch = 300 if instance.speaker_male() else 500
        # x = np.array(instance.audio.get_array_of_samples())
        # f0_swipe = pysptk.swipe(x.astype(np.float64), fs=fs, hopsize=ms_10_in_frames, min=min_pitch, max=max_pitch, otype="f0", threshold=0.27)
        #
        # times = np.arange(0, len(f0_swipe)) / 100.0
        # filtered_times = times[f0_swipe > 0]
        # filtered_f0_swipe = f0_swipe[f0_swipe > 0]
        #
        # return filtered_times, filtered_f0_swipe

    def batch_extract(self, instances):
        return [self.extract(instance) for instance in instances]



# coding: utf-8

import sklearn.cross_validation
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from . import utils


class ClassificationHelper(object):
    def __init__(self, config, classes=[0, 1], classifier=None):
        self.config = config
        self.subsample = config["subsample"]

        if classifier:
            self.classifier = classifier
        else:
            self.classifier = self._build_classifier()
            print(("Using {} classifier".format(self.classifier)))

        self.classes = classes

    def _build_classifier(self):
        classifiers = {
            "SVM": SVC(),
            "RF": RandomForestClassifier(),
            "LDA": LinearDiscriminantAnalysis(),
        }
        clf = classifiers[self.config["classifier_name"]]
        clf.set_params(**self.config["classifier_params"])
        return clf

    def classification_probas(self, X_train, y_train, X_test):
        self.classifier.fit(X_train, y_train)
        return self.classifier.predict_proba(X_test)

    def one_speaker_out_cross_validation(self, X, y, speakers):
        folds = sklearn.cross_validation.LeaveOneLabelOut(speakers)
        return self._cross_validate(folds, X, y)

    def k_fold_cross_validation(self, X, y, n_folds, verbose=True):
        counts = [sum(y == c) for c in self.classes]
        if verbose:
            print(("running classifiers for", list(zip(self.classes, counts))))
        folds = sklearn.cross_validation.StratifiedKFold(y, n_folds=n_folds, random_state=self.config["seed"])
        return self._cross_validate(folds, X, y)

    def _cross_validate(self, folds, X, y):
        fold_results = {}
        n_folds = len(folds)

        for i, (train_index, test_index) in enumerate(folds):
            utils.print_inline("fold: {} of {}".format(i + 1, n_folds))

            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]

            if self.subsample:
                sub_ids = utils.subsample_ids(y_train)
                X_train = X_train[sub_ids]
                y_train = y_train[sub_ids]

            fold_results[i] = self._classify(X_train, X_test, y_train)
            fold_results[i]["actual"] = y_test
            fold_results[i]["y_ids"] = test_index

        return {"categories": self.classes, "fold_results": fold_results, "y": y}

    def _classify(self, X_train, X_test, y_train):
        self.classifier.fit(X_train, y_train)
        y_pred_probabilities = self.classifier.predict_proba(X_test)
        res = {"predicted_probabilities": y_pred_probabilities}

        if self.config["save_weights"]:
            if self.config["classifier_name"] == "RF":
                feature_importances = self.classifier.feature_importances_
            elif "SVM" in self.config["classifier_name"]:
                feature_importances = self.classifier.coef_
            else:
                raise "I don't know whitch weights to save"
            res.update({"classifier_weights": feature_importances})
        return res



# coding: utf-8

import os
import os.path
import subprocess
from time import gmtime, strftime

LOGLEVEL = 3

def mute():
    global LOGLEVEL
    LOGLEVEL = 1

def home():
    return os.path.expanduser("~")


def now():
    return strftime("%Y-%m-%d-%H:%M:%S", gmtime())


def mkdir_p(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def exists(fname):
    return os.path.isfile(fname)


def basename(dir):
    if dir.endswith("/"):
        dir = dir[:-1]
    return os.path.basename(dir)


def cp(src, dst):
    run_external_command("cp", non_named_params=[src, dst])


def rm(filename):
    if exists(filename):
        warning("removing {}".format(filename))
    run_external_command("rm", non_named_params=[filename])


def run_command(cmd, skip=False, verbose=True):
    if verbose:
        if not skip:
            info("(running) {}".format(cmd))
        else:
            warning("(skiping) {}".format(cmd))
    if not skip:
        try:
            return subprocess.check_output(cmd, shell=True).decode("utf-8")
        except subprocess.CalledProcessError as e:
            error(cmd)
            raise e


def run_external_command(cmd, **args):
    if "skip" in args:
        skip = args["skip"]
        del args["skip"]
    else:
        skip = False

    if "non_named_params" in args:
        no_named = " " + " ".join([str(p) for p in args["non_named_params"]])
        del args["non_named_params"]
    else:
        no_named = ""

    for (arg, val) in list(args.items()):
        cmd += " --{} {}".format(arg, val)

    cmd += no_named
    return run_command(cmd, skip=skip)


def run_script(module, **args):
    # Run a script file without loading the environment again (should be imported and have a main function)
    cmd = "./scripts/"
    cmd += module.__name__.replace("scripts.", "")
    cmd += ".py"

    if "skip" in args:
        skip = args["skip"]
        del args["skip"]
    else:
        skip = False

    for (arg, val) in list(args.items()):
        cmd += " --{} '{}'".format(arg, val)

    if skip:
        warning("(skiping) {}".format(cmd))
    else:
        info("(running) {}".format(cmd))
        try:
            return module.main(**args)
        except Exception:
            error(("Running {} with args {}".format(module.__name__, args)))
            raise


def warning(message):
    global LOGLEVEL
    if LOGLEVEL > 1:
        print(('\n (WARNING) \x1b[2;33;40m' + message + '\x1b[0m'))


def error(message):
    global LOGLEVEL
    if LOGLEVEL > 0:
        print(('\n (ERROR) \x1b[2;31;43m' + message + '\x1b[0m'))


def info(message):
    global LOGLEVEL
    if LOGLEVEL > 2:
        print(('\n (INFO) \x1b[2;32;40m' + message + '\x1b[0m'))



from .eeg_feature_extractor import EEGFeatureExtractor


class RawExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(RawExtractor, self).__init__(config)

    def extract(self, trial):
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        ti, tf = self.extraction_time_limits

        initial_frame = int((ti - self.trial_tmin) * self.sfreq)
        end_frame = int((tf - self.trial_tmin) * self.sfreq)
        data = trial[self.channels_to_extract_from, initial_frame:end_frame]
        return data

    def feature_names(self):
        pass
